#pragma comment(lib, "winmm.lib")

#include <ctype.h>
#include <sstream>
#include <iostream>
#include <map>

#include "midi.h"

using std::ostringstream;
using std::cout;
using std::endl;
using std::map;

union MidiMsg
{
   unsigned long word; unsigned char data[4];
};

Midi *Midi::make(int port, string &errMsg)
{
    // MIDI device interface for sending MIDI output

   HMIDIOUT device;

   if(midiOutOpen(&device, port, 0, 0, CALLBACK_NULL) != MMSYSERR_NOERROR)
   {
      ostringstream out;

      out << "Error calling midiOutOpen, port: [" << port << "]";

      errMsg = out.str();

      return(0);
   }

   Midi *m = new Midi(device);

   return(m);
}

Midi::~Midi()
{
   // If any notes are playing, stop them

   midiOutReset(mDevice);

   // Remove any data in MIDI device and close the MIDI Output port

   midiOutClose(mDevice);
}

void Midi::setInstrument(int instrument)
{
   if(mCurrentInstrument == instrument)
   {
      // JFD: What about multiple channels???

      return;
   }

   mCurrentInstrument = instrument;

   MidiMsg msg;

   // ------------------------------------------------------------
   // Our passed-in instrument values will be: 1 <= value <= 128
   // but we'll need to convert to zero based for our
   // call to midiOutShortMsg
   // ------------------------------------------------------------

   msg.data[0] = 0xC0;
   msg.data[1] = mCurrentInstrument - 1;
   msg.data[2] = 0;
   msg.data[3] = 0;

   if(midiOutShortMsg(mDevice, msg.word) != MMSYSERR_NOERROR)
   {
      ostringstream out;

      out <<
         "Error calling midiOutShortMsg, instrument: [" << instrument << "]";

      cout << out.str();
   }
}

bool Midi::doCommand(
   string &errMsg,
   int noteNumber,
   int velocity,
   int midiChannelNumber
)
{
   MidiMsg msg;

   msg.data[0] = 0x90 + midiChannelNumber;
   msg.data[1] = noteNumber;
   msg.data[2] = velocity;
   msg.data[3] = 0;

   if(false)
   {
      printf(
         "%d %d %d - %08x\n",
         midiChannelNumber, noteNumber, velocity, msg.word
      );
   }

   if(midiOutShortMsg(mDevice, msg.word) != MMSYSERR_NOERROR)
   {
      ostringstream out;

      out << "Error calling doCommand("
         << noteNumber << ", " <<
         velocity << ", " << midiChannelNumber << ")";

      errMsg = out.str();

      return(false);
   }

   return(true);
}

void Midi::playNote(
   int noteNumber, int millis, int velocity, int midiChannelNumber
)
{
   note(noteNumber, velocity, midiChannelNumber);
   Sleep(millis);
   note(noteNumber, 0, midiChannelNumber);
}

void Midi::note(int noteNumber, int velocity, int midiChannelNumber)
{
   string errMsg;

   if(!doCommand(errMsg, noteNumber, velocity, midiChannelNumber))
   {
      cout << errMsg << endl;
   }
}

typedef map<string, int> StringToIntMap;
typedef map<int, string> IntToStringMap;

struct VoiceMap
{
   StringToIntMap mNameToNumber;
   IntToStringMap mNumberToName;
};

static VoiceMap *getMap()
{
   static bool initialized = false;
   static VoiceMap theMap;

   if(!initialized)
   {
      static char *list[] =
      {
         // piano

        /* 001 */ "acoustic_grand_piano",
        /* 002 */ "bright_acoustic_piano",
        /* 003 */ "electric_grand_piano",
        /* 004 */ "honky_tonk_piano",
        /* 005 */ "electric_piano_1",
        /* 006 */ "electric_piano_2",
        /* 007 */ "harpsichord",
        /* 008 */ "clavi",

         // chromatic percussion

        /* 009 */ "celesta",
        /* 010 */ "glockenspiel",
        /* 011 */ "music_box",
        /* 012 */ "vibraphone",
        /* 013 */ "marimba",
        /* 014 */ "xylophone",
        /* 015 */ "tubular_bells",
        /* 016 */ "dulcimer",

         // organ

        /* 017 */ "drawbar_organ",
        /* 018 */ "percussive_organ",
        /* 019 */ "rock_organ",
        /* 020 */ "church_organ",
        /* 021 */ "reed_organ",
        /* 022 */ "accordion",
        /* 023 */ "harmonica",
        /* 024 */ "tango_accordion",

         // guitar

        /* 025 */ "acoustic_guitar_nylon",
        /* 026 */ "acoustic_guitar_steel",
        /* 027 */ "electric_guitar_jazz",
        /* 028 */ "electric_guitar_clean",
        /* 029 */ "electric_guitar_muted",
        /* 030 */ "overdriven_guitar",
        /* 031 */ "distortion_guitar",
        /* 032 */ "guitar_harmonics",

         // bass

        /* 033 */ "acoustic_bass",
        /* 034 */ "electric_bass_finger",
        /* 035 */ "electric_bass_pick",
        /* 036 */ "fretless_bass",
        /* 037 */ "slap_bass_1",
        /* 038 */ "slap_bass_2",
        /* 039 */ "synth_bass_1",
        /* 040 */ "synth_bass_2",

         // solo strings

        /* 041 */ "violin",
        /* 042 */ "viola",
        /* 043 */ "cello",
        /* 044 */ "contrabass",
        /* 045 */ "tremolo_strings",
        /* 046 */ "pizzicato_strings",
        /* 047 */ "orchestral_harp",
        /* 048 */ "timpani",

         // ensemble

        /* 049 */ "string_ensemble_1",
        /* 050 */ "string_ensemble_2",
        /* 051 */ "synthstrings_1",
        /* 052 */ "synthstrings_2",
        /* 053 */ "choir_aahs",
        /* 054 */ "voice_oohs",
        /* 055 */ "synth_voice",
        /* 056 */ "orchestra_hit",

         // brass

        /* 057 */ "trumpet",
        /* 058 */ "trombone",
        /* 059 */ "tuba",
        /* 060 */ "muted_trumpet",
        /* 061 */ "french_horn",
        /* 062 */ "brass_section",
        /* 063 */ "synthbrass_1",
        /* 064 */ "synthbrass_2",

         // reed

        /* 065 */ "soprano_sax",
        /* 066 */ "alto_sax",
        /* 067 */ "tenor_sax",
        /* 068 */ "baritone_sax",
        /* 069 */ "oboe",
        /* 070 */ "english_horn",
        /* 071 */ "bassoon",
        /* 072 */ "clarinet",

         // pipe

        /* 073 */ "piccolo",
        /* 074 */ "flute",
        /* 075 */ "recorder",
        /* 076 */ "pan_flute",
        /* 077 */ "blown_bottle",
        /* 078 */ "shakuhachi",
        /* 079 */ "whistle",
        /* 080 */ "ocarina",

         // synth lead

        /* 081 */ "lead_1_square",
        /* 082 */ "lead_2_sawtooth",
        /* 083 */ "lead_3_calliope",
        /* 084 */ "lead_4_chiff",
        /* 085 */ "lead_5_charang",
        /* 086 */ "lead_6_voice",
        /* 087 */ "lead_7_fifths",
        /* 088 */ "lead_8_bass_lead",

         // synth pad

        /* 089 */ "pad_1_new_age",
        /* 090 */ "pad_2_warm",
        /* 091 */ "pad_3_polysynth",
        /* 092 */ "pad_4_choir",
        /* 093 */ "pad_5_bowed",
        /* 094 */ "pad_6_metallic",
        /* 095 */ "pad_7_halo",
        /* 096 */ "pad_8_sweep",

         // synth effects

        /* 097 */ "fx_1_rain",
        /* 098 */ "fx_2_soundtrack",
        /* 099 */ "fx_3_crystal",
        /* 100 */ "fx_4_atmosphere",
        /* 101 */ "fx_5_brightness",
        /* 102 */ "fx_6_goblins",
        /* 103 */ "fx_7_echoes",
        /* 104 */ "fx_8_sci_fi",

         // ethnic

        /* 105 */ "sitar",
        /* 106 */ "banjo",
        /* 107 */ "shamisen",
        /* 108 */ "koto",
        /* 109 */ "kalimba",
        /* 110 */ "bag_pipe",
        /* 111 */ "fiddle",
        /* 112 */ "shanai",

         // percussive

        /* 113 */ "tinkle_bell",
        /* 114 */ "agogo",
        /* 115 */ "steel_drums",
        /* 116 */ "woodblock",
        /* 117 */ "taiko_drum",
        /* 118 */ "melodic_tom",
        /* 119 */ "synth_drum",
        /* 120 */ "reverse_cymbal",

         // sound effects

        /* 121 */ "guitar_fret_noise",
        /* 122 */ "breath_noise",
        /* 123 */ "seashore",
        /* 124 */ "bird_tweet",
        /* 125 */ "telephone_ring",
        /* 126 */ "helicopter",
        /* 127 */ "applause",
        /* 128 */ "gunshot"
      };

      int n = sizeof(list) / sizeof(list[0]);

      for(int i = 0; i < n; i ++)
      {
         theMap.mNumberToName[i + 1]   = list[i];
         theMap.mNameToNumber[list[i]] = i + 1;
      }

      initialized = true;
   }

   return(&theMap);
}

int Midi::instrumentNameToNumber(const string &name)
{
   VoiceMap *map = getMap();

   string lower = "";

   for(int n = name.size(), i = 0; i < n; i ++)
   {
      char c = name[i];
      lower += (isupper(c) ? tolower(c) : c);
   }

   StringToIntMap::iterator it = map->mNameToNumber.find(lower);

   if(it == map->mNameToNumber.end())
      return(-1);

   return(it->second);
}

string Midi::instrumentNumberToName(int number)
{
   VoiceMap *map = getMap();

   IntToStringMap::iterator it = map->mNumberToName.find(number);

   if(it == map->mNumberToName.end())
      return("");

   return(it->second);
}

Midi::Midi(HMIDIOUT device) : mDevice(device), mCurrentInstrument(-1)
{
}
