/*****************************************************************************
 * C/C++ Headers
 *****************************************************************************/

#include <stdio.h>
#include <ctype.h>

#include <algorithm>
#include <iostream>
#include <sstream>
#include <vector>
#include <string>
#include <map>

#include "midi.h"
#include "cleaner.h"
#include "parser.h"

/*****************************************************************************
 * Using Declarations
 *****************************************************************************/

using std::ostringstream;
using std::vector;
using std::string;
using std::cout;
using std::endl;
using std::map;
using std::sort;

/*****************************************************************************
 * Defines
 *****************************************************************************/

#define FOREACH(type, object, name) \
   for(type::iterator name = object.begin(); name != object.end(); name ++)

#define COMMENT_CHAR '/'

// ---------------------
// struct Token
// ---------------------

struct Token
{
   string mTxt;
   int    mLineNumber;

   Token(const string &txt, int lineNumber) :
      mTxt(txt), mLineNumber(lineNumber)
   {
   }
};

typedef vector<Token> TokenVector;

// ---------------------
// class SleepEvent
// ---------------------

class SleepEvent : public Event
{
   public:
      SleepEvent(int duration) : mDuration(duration) {}
      string toString();
      void apply(Midi *m);

   private:
      int mDuration;
};

string SleepEvent::toString()
{
   ostringstream out;
   out << "sleep   :" << mDuration;
   return(out.str());
}

void SleepEvent::apply(Midi *)
{
   Sleep(mDuration);
}

// ---------------------
// class NoteOnEvent
// ---------------------

class NoteOnEvent : public Event
{
   public:
      NoteOnEvent(int number, int volume, int instrument) :
         mNumber(number), mVolume(volume), mInstrument(instrument)
      {
      }

      string toString();
      void apply(Midi *m);

   private:
      int mNumber;
      int mVolume;
      int mInstrument;
};

string NoteOnEvent::toString()
{
   ostringstream out;

   out << "note_on :" << mNumber << ", " << mVolume << ", " << mInstrument;

   return(out.str());
}

void NoteOnEvent::apply(Midi *m)
{
   m->setInstrument(mInstrument);
   m->note(mNumber, mVolume, 0);
}

// ---------------------
// class NoteOffEvent
// ---------------------

class NoteOffEvent : public Event
{
   public:
      NoteOffEvent(int number) : mNumber(number) { }
      string toString();
      void apply(Midi *m);

   private:
      int mNumber;
};

string NoteOffEvent::toString()
{
   ostringstream out;
   out << "note_off:" << mNumber;
   return(out.str());
}

void NoteOffEvent::apply(Midi *m)
{
   m->note(mNumber, 0, 0);
}

// ---------------------
// class NoteStateChange
// ---------------------

class NoteStateChange
{
   public:
      NoteStateChange(
         int timeOffset,
         int number,
         int volume     = 0,
         int instrument = 0
      ) :
         mTimeOffset(timeOffset),
         mNumber(number),
         mVolume(volume),
         mInstrument(instrument)
      {
      }

      bool isOn()
      {
         return(mVolume > 0 && mInstrument > 0);
      }

      int getTimeOffset() { return(mTimeOffset); }
      int getNumber()     { return(mNumber);     }
      int getVolume()     { return(mVolume);     }
      int getInstrument() { return(mInstrument); }

   private:
      int mTimeOffset;
      int mNumber;
      int mVolume;
      int mInstrument;
};

typedef vector<NoteStateChange> NoteStateChangeVector;

// ---------------------
// class Node
// ---------------------

class Node
{
   public:
      virtual ~Node()
      {
         if(false)
         {
            cout << "Whack: " << this << endl;
         }
      }
      virtual int    getDuration()        = 0;
      virtual string toString(int indent) = 0;
      virtual Node   *clone()             = 0;
      virtual void   shift(int offset)    = 0;

      virtual void doNoteStateChanges(
         NoteStateChangeVector &changes, int &timeOffset
      )
      {
      }

      string toString()
      {
         return(toString(0));
      }

      static string makeIndent(int indent)
      {
         string s = "";

         for(int i = 0; i < indent; i ++)
            s += " ";

         return(s);
      }
};

typedef vector<Node *>        NodePtrVector;
typedef vector<NodePtrVector> NodePtrVectorVector;

// ---------------------
// class SilenceNode
// ---------------------

class SilenceNode : public Node
{
   public:

      SilenceNode(int duration);

      int getDuration();

      string toString(int indent);

      Node *clone();
      void shift(int offset);

      void SilenceNode::doNoteStateChanges(
         NoteStateChangeVector &changes, int &timeOffset
      );

   private:

      int mDuration;
};

SilenceNode::SilenceNode(int duration) : mDuration(duration)
{
}

int SilenceNode::getDuration()
{
   return(mDuration);
}

void SilenceNode::doNoteStateChanges(
   NoteStateChangeVector &changes, int &timeOffset
)
{
   timeOffset += getDuration();
}

Node *SilenceNode::clone()
{
   return(new SilenceNode(getDuration()));
}

void SilenceNode::shift(int)
{
}

string SilenceNode::toString(int indent)
{
   ostringstream out;

   out <<
      Node::makeIndent(indent) <<
      "Silence(duration: " << mDuration << ")";

   return(out.str());
}

// ---------------------
// class NoteNode
// ---------------------

class NoteNode : public Node
{
   public:

      NoteNode(int noteNumber, int duration, int volume, int instrument);

      int getDuration();

      string toString(int indent);

      Node *clone();
      void shift(int offset);

      void doNoteStateChanges(
         NoteStateChangeVector &changes, int &timeOffset
      );

      int getNoteNumber();
      int getVolume();
      int getInstrument();

   private:

      int mNoteNumber;
      int mDuration;
      int mVolume;
      int mInstrument;
};

NoteNode::NoteNode(
   int noteNumber, int duration, int volume, int instrument
) :
   mNoteNumber(noteNumber),
   mDuration(duration),
   mVolume(volume),
   mInstrument(instrument)
{
}

int NoteNode::getDuration()
{
   return(mDuration);
}

string NoteNode::toString(int indent)
{
   ostringstream out;

   out <<
      Node::makeIndent(indent) <<
      "Note(number: " << mNoteNumber <<", duration: " <<
      mDuration << ", volume: " << mVolume << ", instrument: " <<
      mInstrument << ")";

   return(out.str());
}

Node *NoteNode::clone()
{
   return(
      new NoteNode(
         getNoteNumber(),
         getDuration(),
         getVolume(),
         getInstrument()
      )
   );
}

void NoteNode::shift(int offset)
{
   int newNoteNumber = getNoteNumber() + offset;

   if(newNoteNumber >= MIDI_MIN_NOTE && newNoteNumber <= MIDI_MAX_NOTE)
      mNoteNumber = newNoteNumber;
}

void NoteNode::doNoteStateChanges(
   NoteStateChangeVector &changes, int &timeOffset
)
{
   changes.push_back(
      NoteStateChange(timeOffset, mNoteNumber, mVolume, mInstrument)
   );

   timeOffset += getDuration();

   changes.push_back(NoteStateChange(timeOffset, mNoteNumber));
}

int NoteNode::getNoteNumber() { return(mNoteNumber); }
int NoteNode::getVolume()     { return(mVolume);     }
int NoteNode::getInstrument() { return(mInstrument); }

// ---------------------
// class GroupNode
// ---------------------

class GroupNode : public Node
{
   public:

      GroupNode(NodePtrVectorVector groups);

      ~GroupNode();

      int getDuration();

      string toString(int indent);

      Node *clone();

      void shift(int offset);

      void doNoteStateChanges(
         NoteStateChangeVector &changes, int &timeOffset
      );
 
      NodePtrVectorVector get();

   private:

      NodePtrVectorVector mGroups;
};

GroupNode::GroupNode(NodePtrVectorVector groups) : mGroups(groups)
{
}

GroupNode::~GroupNode()
{
   for(int n = mGroups.size(), i = 0; i < n; i ++)
   {
      NodePtrVector npv = mGroups[i];

      for(int nn = mGroups[i].size(), j = 0; j < nn; j ++)
      {
         delete(mGroups[i][j]);
         mGroups[i][j] = NULL;
      }

      mGroups[i].clear();
   }

   mGroups.clear();
}

int GroupNode::getDuration()
{
   int value = 0;

   NodePtrVectorVector::iterator outer = mGroups.begin();

   while(outer != mGroups.end())
   {
      NodePtrVector::iterator inner = outer->begin();
      int sum                       = 0;

      while(inner != outer->end())
      {
         sum += (*inner)->getDuration();
         inner++;
      }

      outer ++;

      if(sum > value)
         value = sum;
   }

   return(value);
}

string GroupNode::toString(int indent)
{
   ostringstream out;

   string i1 = Node::makeIndent(indent);
   string i2 = Node::makeIndent(indent + 1);

   out << i1 << "Group(" << endl;

   for(int n = mGroups.size(), i = 0; i < n; i ++)
   {
      out << i2 << "(" << endl;

      for(int nn = mGroups[i].size(), j = 0; j < nn; j ++)
      {
         out << mGroups[i][j]->toString(indent + 2) << endl;
      }

      out << i2 << ")" << endl;
   }

   out << i1 << ")";

   return(out.str());
}

Node *GroupNode::clone()
{
   NodePtrVectorVector groups;

   for(int n = mGroups.size(), i = 0; i < n; i ++)
   {
      NodePtrVector group;

      for(int nn = mGroups[i].size(), j = 0; j < nn; j ++)
      {
         group.push_back(mGroups[i][j]->clone());
      }

      groups.push_back(group);
   }

   return(new GroupNode(groups));
}

void GroupNode::shift(int offset)
{
   for(int n = mGroups.size(), i = 0; i < n; i ++)
   {
      for(int nn = mGroups[i].size(), j = 0; j < nn; j ++)
      {
         mGroups[i][j]->shift(offset);
      }
   }
}

void GroupNode::doNoteStateChanges(
   NoteStateChangeVector &changes, int &timeOffset
)
{
   NodePtrVectorVector::iterator outer = mGroups.begin();

   while(outer != mGroups.end())
   {
      NodePtrVector::iterator inner = outer->begin();
      int offset                    = timeOffset;

      while(inner != outer->end())
      {
         (*inner)->doNoteStateChanges(changes, offset);
         inner++;
      }

      outer++;
   }

   timeOffset += getDuration();
}

// ---------------------
// struct GroupEntry
// ---------------------

struct GroupEntry
{
   int mLPIndex;
   int mRPIndex;

   GroupEntry(int lpIndex, int rpIndex) :
      mLPIndex(lpIndex), mRPIndex(rpIndex)
   {
   }
};

typedef vector<GroupEntry> GroupEntryVector;

struct GroupDelimiters
{
   bool mDelimiterFound;
   string mLeft;
   string mRight;

   GroupDelimiters() :
      mDelimiterFound(false), mLeft("("), mRight(")")
   {
   }
};

/*****************************************************************************
 * Forward Declarations Required By Class: Parser
 *****************************************************************************/

static string toLowerCase(const string &in);

static bool getInstrumentNumber(const string &s, int *value, string &errMsg);

static bool getVolume(const string &s, int *value, string &errMsg);

static TokenVector stringToTokens(
   const string &txt, GroupDelimiters &groupDelimiters
);

static void dumpTokens(TokenVector &tokens);

static Node *getNode(
   const TokenVector &t,
   int               &idx,
   GroupDelimiters   &groupDelimiters,
   int               defaultVolume,
   int               defaultInstrument,
   string            &errMsg
);

static void dumpNodes(NodePtrVector &nodes);

static void dumpChanges(string caption, NoteStateChangeVector &changes);

static void dumpEvents(EventPtrVector &events);

// -------------------------------
// main class: Parser
// -------------------------------

bool Parser::getArgs(
   int    argc,
   char   **argv,
   int    *defaultVolumePtr,
   int    *defaultInstrumentPtr,
   string &fileName,
   string &errMsg
)
{
   int i            = 1;
   int nFiles       = 0;
   int nVolumes     = 0;
   int nInstruments = 0;

   while(i < argc)
   {
      string theArg     = toLowerCase(argv[i]);
      bool isInstrument = (theArg == PARSER_INSTRUMENT);
      bool isVolume     = (theArg == PARSER_VOLUME);

      if(isInstrument || isVolume)
      {
         if(i + 1 >= argc)
         {
            errMsg = (isInstrument ? "Instrument" : "Volume");
            errMsg += " exected after: '";
            errMsg += argv[i];
            errMsg += "'";

            return(false);
         }

         const char *nextArg = argv[i + 1];

         if(isInstrument)
         {
            if(!getInstrumentNumber(nextArg, defaultInstrumentPtr, errMsg))
               return(false);

            nInstruments++;
         }
         else
         {
            if(!getVolume(nextArg, defaultVolumePtr, errMsg))
               return(false);

            nVolumes++;
         }

         i++;
      }
      else
      {
         fileName = argv[i];
         nFiles++;
      }

      i++;
   }

   if(nFiles == 0)
   {
      errMsg = "No fileName supplied";
      return(false);
   }

   if(nFiles > 1)
   {
      errMsg = "Multiple fileNames supplied";
      return(false);
   }

   if(nVolumes > 1)
   {
      errMsg = "Multiple volumes supplied";
      return(false);
   }

   if(nInstruments > 1)
   {
      errMsg = "Multiple instruments supplied";
      return(false);
   }

   if(true)
   {
      cout <<
         "defaultVolume: ["     << *defaultVolumePtr     << "], " <<
         "defaultInstrument: [" << *defaultInstrumentPtr << "], " <<
         "fileName: ["          << fileName              << "]"   << endl;
   }

   return(true);
}

bool Parser::fileToString(
   const string &fileName, string &contents, string &errMsg
)
{
   FILE *f = NULL;

   if((f = fopen(fileName.c_str(), "rb")) == NULL)
   {
      errMsg = "Could not open file: [";
      errMsg += fileName;
      errMsg += "]";

      return(false);
   }

   int c    = 0;
   contents = "";

   while((c = fgetc(f)) != EOF)
      contents += (char)c;

   fclose(f);

   return(true);
}

bool Parser::getEvents(
   const          string &contents,
   int            defaultVolume,
   int            defaultInstrument,
   EventPtrVector &events,
   int            &timeOffset,
   string         &errMsg
)
{
   // Attempt to break 'contents' down into tokens

   GroupDelimiters groupDelimiters;

   TokenVector tokens = stringToTokens(contents, groupDelimiters);
   int nTokens        = tokens.size();

   if(false)
      dumpTokens(tokens);

   // Now convert our tokens into a series of Node* objects

   NodePtrVector nodes;
   Cleaner       theCleaner;

   int idx = 0;

   while(idx < nTokens)
   {
      Node *n = getNode(
         tokens,
         idx,
         groupDelimiters,
         defaultVolume,
         defaultInstrument,
         errMsg
      );

      if(!n)
         return(false);

      nodes.push_back(n);

      // Ensure that 'n' will always be deleted

      theCleaner.addPointer(n);

      if(false)
      {
         cout << "Added Node: [" << n->toString() << "]" << endl;
      }
   }

   int nNodes = nodes.size();

   if(nNodes == 0)
   {
      errMsg = "No commands found";
      return(false);
   }

   if(false)
      dumpNodes(nodes);

   // Now, convert our nodes to a series of state change
   // (note on / note off) events

   NoteStateChangeVector changes;

   for(int i = 0; i < nNodes; i ++)
      nodes[i]->doNoteStateChanges(changes, timeOffset);

   // Now sort our state changes by time offset

   if(false)
      dumpChanges("Changes - Before", changes);

   sort(changes.begin(), changes.end());

   if(false)
      dumpChanges("Changes - After", changes);

   // Now, convert to events ...

   int lastTimeOffset = 0;
   int cSize          = changes.size();

   for(int j = 0; j < cSize; j ++)
   {
      int d = changes[j].getTimeOffset();

      if(j == 0)
      {
         if(d > 0)
            events.push_back(new SleepEvent(d));
      }

      if(changes[j].isOn())
      {
         events.push_back(
            new NoteOnEvent(
               changes[j].getNumber(),
               changes[j].getVolume(),
               changes[j].getInstrument()
            )
         );
      }
      else
      {
         events.push_back(
            new NoteOffEvent(changes[j].getNumber())
         );
      }

      if(j + 1 < cSize)
      {
         int delta = (changes[j + 1].getTimeOffset() - d);

         if(delta > 0)
            events.push_back(new SleepEvent(delta));
      }
      else
      {
         lastTimeOffset = d;
      }
   }

   // Handle trailing silence

   if(timeOffset > lastTimeOffset)
      events.push_back(new SleepEvent(timeOffset - lastTimeOffset));

   if(false)
      dumpEvents(events);

   return(true);
}

/*****************************************************************************
 * Local Functions
 *****************************************************************************/

static string toLowerCase(const string &in)
{
   string lower = "";

   for(int n = in.size(), i = 0; i < n; i ++)
   {
      char c = in[i];

      lower += (isupper(c) ? tolower(c) : c);
   }

   return(lower);
}

static void handleParentheses(
   TokenVector     &tokens,
   const string    &txt,
   GroupDelimiters &groupDelimiters,
   int             lineNumber
)
{
   static string delimiterEntries[][2] =
   {
      { "(", ")" },
      { "{", "}" },
      { "<", ">" },
      { "[", "]" }
   };

   string current = "";
   bool inToken   = false;

   for(int n = txt.size(), i = 0; i < n; i ++)
   {
      string c = txt.substr(i, 1);

      if(!groupDelimiters.mDelimiterFound)
      {
         bool found = false;
         int ne     = sizeof(delimiterEntries) / sizeof(delimiterEntries[0]);

         for(int j = 0; !groupDelimiters.mDelimiterFound && j < ne; j ++)
         {
            string d0 = delimiterEntries[j][0];
            string d1 = delimiterEntries[j][1];

            if(c == d0 || c == d1)
            {
               groupDelimiters.mLeft           = d0;
               groupDelimiters.mRight          = d1;
               groupDelimiters.mDelimiterFound = true;
            }
         }
      }

      if(c == groupDelimiters.mLeft || c == groupDelimiters.mRight)
      {
         if(current != "")
         {
            tokens.push_back(Token(toLowerCase(current), lineNumber));
            current = "";
         }

         tokens.push_back(Token(c, lineNumber));
      }
      else
      {
         current += c;
      }
   }

   if(current != "")
   {
      tokens.push_back(Token(toLowerCase(current), lineNumber));
   }
}

static TokenVector stringToTokens(
   const string &txt, GroupDelimiters &groupDelimiters
)
{
   typedef enum { STATE_DEFAULT, STATE_TOKEN, STATE_COMMENT } State;

   TokenVector tokens;

   State currentState = STATE_DEFAULT;
   int n              = txt.size();
   int lineNumber     = 1;
   string current     = "";

   for(int i = 0; i < n; i ++)
   {
      char c = txt[i];

      switch(currentState)
      {
         case STATE_TOKEN:
         {
            if(isspace(c) || c == COMMENT_CHAR || i + 1 == n)
            {
               handleParentheses(
                  tokens, current, groupDelimiters, lineNumber
               );

               current = "";
               currentState = STATE_DEFAULT;
            }
            else
            {
               current += c;
            }
         }
         break;

         case STATE_COMMENT:
         {
            if(c == '\n' || i + 1 == n)
            {
               currentState = STATE_DEFAULT;
            }
         }
         break;

         default:
         {
            if(!isspace(c))
            {
               if(c == COMMENT_CHAR)
               {
                  currentState = STATE_COMMENT;
               }
               else
               {
                  current     += c;
                  currentState = STATE_TOKEN;
               }
            }
         }
         break;
      }

      if(c == '\n')
         lineNumber++;
   }

   return(tokens);
}

static bool isInteger(const string &s)
{
   for(int n = s.size(), i = 0; i < n; i ++)
      if(!isdigit(s[i]))
         return(false);

   return(true);
}

static int toPositiveInteger(const string &s)
{
   if(!isInteger(s))
      return(-1);

   int value = atoi(s.c_str());

   return(value == 0 ? -1 : value);
}

typedef map<string, int> StringToIntMap;

static StringToIntMap *getMap()
{
   static bool initialized = false;
   static StringToIntMap theMap;

   static const char *notes = "ccddeffggaab";
   static const char *sf    = " x x  x x x ";

   if(!initialized)
   {
      int noteNumber = 0;

      for(int i = -1; i <= 9; i ++)
      {
         for(int j = 0; j < 12; j ++, noteNumber ++)
         {
            string s      = string("") + notes[j];
            string suffix = "";

            if(i < 0)
            {
               suffix = "m1";
            }
            else
            {
               suffix += (char)(i + '0');
            }

            if(sf[j] == 'x')
            {
               theMap[s + "#" + suffix] = noteNumber;

               theMap[string("") + notes[(j + 1) % 12] + "b" + suffix] =
                  noteNumber;
            }
            else
            {
               theMap[s + suffix] = noteNumber;
            }

            if(noteNumber >= MIDI_MAX_NOTE)
               break;
         }
      }

      if(false)
      {
         FOREACH(StringToIntMap, theMap, it)
            cout << it->first << " => " << it->second << endl;
      }

      initialized = true;
   }

   return(&theMap);
}

static int noteNameToNumber(const string &s)
{
   StringToIntMap *map         = getMap();
   StringToIntMap::iterator it = map->find(s);

   if(it == map->end())
      return(-1);

   return(it->second);
}

static bool getItem(
   const  string &s,
   const  string &which,
   int    (*nameToNumberFunction)(const string &),
   int    minValue,
   int    maxValue,
   int    *value,
   string &errMsg
)
{
   int theValue = 0;
   errMsg       = "";

   ostringstream out;

   out << "Invalid " << which << ": [" << s << "]";

   if((theValue = toPositiveInteger(s)) != -1)
   {
      if(theValue < minValue || theValue > maxValue)
      {
         out <<
            ", " << "expected: " <<
            minValue << " <= " << which << " <= " << maxValue;

         errMsg = out.str();
      }
   }
   else if((theValue = nameToNumberFunction(s)) == -1)
   {
      errMsg = out.str();
   }

   if(errMsg != "")
      return(false);

   *value = theValue;

   return(true);
}

static bool getInstrumentNumber(const string &s, int *value, string &errMsg)
{
   return(
      getItem(
         s,
         "instrument",
         Midi::instrumentNameToNumber,
         MIDI_MIN_INSTRUMENT,
         MIDI_MAX_INSTRUMENT,
         value,
         errMsg
      )
   );
}

static bool getNoteNumber(const string &s, int *value, string &errMsg)
{
   return(
      getItem(
         s,
         "note",
         noteNameToNumber,
         MIDI_MIN_NOTE,
         MIDI_MAX_NOTE,
         value,
         errMsg
      )
   );
}

static bool getVolume(const string &s, int *value, string &errMsg)
{
   ostringstream out;

   int iValue = toPositiveInteger(s);

   if(iValue == -1 || iValue > MIDI_MAX_VOLUME)
   {
      out <<
         "Invalid volume: [" << s << "], expected: " <<
         MIDI_MIN_VALUE << " <= volume <= " << MIDI_MAX_VALUE;

      errMsg = out.str();

      return(false);
   }

   *value = iValue;

   return(true);
}

static Node *getNode(
   const TokenVector &t,
   int               &idx,
   GroupDelimiters   &groupDelimiters,
   int               defaultVolume,
   int               defaultInstrument,
   string            &errMsg
)
{
   ostringstream out;

   Token theToken = t[idx];
   int n          = t.size();

   if(theToken.mTxt == PARSER_SILENCE)
   {
      ++idx;

      if(idx >= n)
      {
         out << "Expected duration after: 'x', line: " <<
            "[" << theToken.mLineNumber << "]";

         errMsg = out.str();

         return(NULL);
      }

      theToken     = t[idx];
      int duration = 0;

      if((duration = toPositiveInteger(theToken.mTxt)) == -1)
      {
         out <<
            "Invalid duration after 'x': [" << theToken.mTxt << "], " <<
            "line: [" << theToken.mLineNumber << "]";

         errMsg = out.str();

         return(NULL);
      }

      ++idx;

      return(new SilenceNode(duration));
   }

   if(theToken.mTxt == PARSER_REPEAT)
   {
      // A repeat command, should be followed by a positive integer

      ++idx;

      if(idx >= n)
      {
         out << "Expected number after: '" << PARSER_REPEAT << "', line: " <<
            "[" << theToken.mLineNumber << "]";

         errMsg = out.str();

         return(NULL);
      }

      theToken   = t[idx];
      int rValue = 0;

      if((rValue = toPositiveInteger(theToken.mTxt)) == -1)
      {
         out <<
            "Invalid value after '" << PARSER_REPEAT <<
            "': [" << theToken.mTxt << "], " <<
            "line: [" << theToken.mLineNumber << "]";

         errMsg = out.str();

         return(NULL);
      }

      ++idx;

      if(idx >= n)
      {
         out <<
            "Expected command after: '" << PARSER_REPEAT << "' " <<
            rValue << ", line: " << "[" << theToken.mLineNumber << "]";

         errMsg = out.str();

         return(NULL);
      }

      Node *theNode = getNode(
         t, idx, groupDelimiters, defaultVolume, defaultInstrument, errMsg
      );

      if(!theNode)
         return(NULL);

      // Now, we create a GroupNode with theNode 
      // repeated rValue times

      NodePtrVectorVector groups;
      NodePtrVector       group;

      for(int i = 0; i < rValue; i ++)
      {
         if(i == 0)
            group.push_back(theNode);

         else
            group.push_back(theNode->clone());
      }

      groups.push_back(group);

      return(new GroupNode(groups));
   }

   if(theToken.mTxt == PARSER_SHIFT)
   {
      // A shift command, should be followed by an integer (+ or -)

      ++idx;

      if(idx >= n)
      {
         out <<
            "Expected step offset after: '" << PARSER_SHIFT << "', line: " <<
            "[" << theToken.mLineNumber << "]";

         errMsg = out.str();

         return(NULL);
      }

      theToken           = t[idx];
      string theString   = theToken.mTxt;
      string firstLetter = theString.substr(0, 1);
      bool isNegative    = (firstLetter == "-");

      if(firstLetter == "+" || isNegative)
         theString = theString.substr(1);

      // Remainder of 'theString' should be an integer - the
      // number of steps to shift

      if(!isInteger(theString))
      {
         out <<
            "Invalid value: [" << theToken.mTxt << "], after: '" <<
            PARSER_SHIFT << "', line: [" << theToken.mLineNumber << "]";

         errMsg = out.str();

         return(NULL);
      }

      int theOffset = (isNegative ? -1 : 1) * atoi(theString.c_str());

      ++idx;

      if(idx >= n)
      {
         out <<
            "Expected command after: '" << PARSER_SHIFT << "' " <<
            theOffset << ", line: " << "[" << theToken.mLineNumber << "]";

         errMsg = out.str();

         return(NULL);
      }

      Node *theNode = getNode(
         t,
         idx,
         groupDelimiters,
         defaultVolume,
         defaultInstrument,
         errMsg
      );

      if(!theNode)
         return(NULL);

      theNode->shift(theOffset);

      return(theNode);
   }

   if(theToken.mTxt == groupDelimiters.mLeft)
   {
      GroupEntryVector geVector;
      bool first = true;

      while(true)
      {
         // If we've run out of tokens prematurely ...

         idx++;

         if(idx == n)
         {
            if(first)
            {
               out <<
                  "Expected '" << groupDelimiters.mLeft << "' after '" <<
                  groupDelimiters.mLeft << "', line: [" <<
                  theToken.mLineNumber << "]";
            }
            else
            {
               out <<
                  "Expected '" << groupDelimiters.mRight << "' to match '" <<
                  groupDelimiters.mLeft << "', line: [" <<
                  theToken.mLineNumber << "]";
            }

            errMsg = out.str();

            return(NULL);
         }

         // Now, we expect a left or right delimiter,
         // anything else is an error

         if(t[idx].mTxt == groupDelimiters.mLeft)
         {
            int lpIndex = idx;
            int rpIndex = -1;
            int lpCount = 1;
            int rpCount = 0;

            // Try to find the matching right for this left

            for(int i = lpIndex + 1; i < n; i ++)
            {
               if(t[i].mTxt == groupDelimiters.mLeft)
               {
                  lpCount++;
               }
               else if(t[i].mTxt == groupDelimiters.mRight)
               {
                  rpCount++;

                  if(rpCount == lpCount)
                  {
                     rpIndex = i;
                     break;
                  }
               }
            }

            if(rpIndex == -1)
            {
               out <<
                  "No matching '" << groupDelimiters.mRight << "' for '" <<
                  groupDelimiters.mLeft << ", line: [" <<
                  t[idx].mLineNumber << "]";

               errMsg = out.str();

               return(NULL);
            }

            // Store off the starting/ending token numbers
            // for this group, after we leave the while(true)
            // loop we'll produce a GroupNode from the
            // groups we found

            geVector.push_back(GroupEntry(lpIndex, rpIndex));

            idx = rpIndex;
         }
         else if(t[idx].mTxt == groupDelimiters.mRight)
         {
            idx++;
            break;
         }
         else
         {
            out <<
               "Expected '" << groupDelimiters.mRight << "' or '" <<
               groupDelimiters.mLeft << "' not: [" <<
               t[idx].mTxt << "], line: [" << t[idx].mLineNumber << "]";

            errMsg = out.str();

            return(NULL);
         }

         if(first)
            first = false;
      }

      int nGEV = geVector.size();

      if(nGEV == 0)
      {
         out << "Empty group, line: [" << theToken.mLineNumber << "]";
         errMsg = out.str();
         return(false);
      }

      // Create a GroupNode from the groups we encountered

      NodePtrVectorVector groups;

      for(int j = 0; j < nGEV; j ++)
      {
         int theIDX = geVector[j].mLPIndex + 1;

         NodePtrVector group;

         while(theIDX < geVector[j].mRPIndex)
         {
            Node *theNode = getNode(
               t,
               theIDX,
               groupDelimiters,
               defaultVolume,
               defaultInstrument,
               errMsg
            );

            if(!theNode)
               return(NULL);

            group.push_back(theNode);
         }

         groups.push_back(group);
      }

      return(new GroupNode(groups));
   }

   // Only option left - a note - can be
   // a mnemonic (i.e. f#4) or a midi note number

   int noteNumber = 0;

   if(!getNoteNumber(theToken.mTxt, &noteNumber, errMsg))
   {
      out << errMsg << ", line: [" << theToken.mLineNumber << "]";
      errMsg = out.str();
      return(NULL);
   }

   ++idx;

   if(idx >= n)
   {
      out << "Expected duration after note, line: " <<
         "[" << theToken.mLineNumber << "]";

      errMsg = out.str();
      return(NULL);
   }

   theToken     = t[idx];
   int duration = 0;

   if((duration = toPositiveInteger(theToken.mTxt)) == -1)
   {
      out <<
         "Invalid duration: [" << theToken.mTxt << "] after note, line: " <<
         "[" << theToken.mLineNumber << "], " << errMsg;

      errMsg = out.str();
      return(NULL);
   }

   ++idx;

   // If we've run out of tokens or if the next token
   // isn't part of this note (i.e. instrument or volume)
   // then we're done building this NoteNode

   if(idx >= n ||
      (t[idx].mTxt != PARSER_VOLUME && t[idx].mTxt != PARSER_INSTRUMENT))
   {
      return(
         new NoteNode(
            noteNumber, duration, defaultVolume, defaultInstrument
         )
      );
   }

   theToken = t[idx];
   idx++;

   bool foundVolume     = (theToken.mTxt == PARSER_VOLUME);
   bool foundInstrument = !foundVolume;
   int volume           = defaultVolume;
   int instrument       = defaultInstrument;

   if(idx >= n)
   {
      out <<
         "Expected " <<
         (foundVolume ? "volume" : "instrument")
         << " after '" <<
         (foundVolume ? PARSER_VOLUME : PARSER_INSTRUMENT) << 
         "', line: [" << theToken.mLineNumber << "]";

      errMsg = out.str();
      return(NULL);
   }

   theToken   = t[idx];
   bool error = false;

   if(foundVolume)
   {
      if(!getVolume(theToken.mTxt, &volume, errMsg))
         error = true;
   }
   else
   {
      if(!getInstrumentNumber(theToken.mTxt, &instrument, errMsg))
         error = true;
   }

   if(error)
   {
      out << errMsg << ", line: [" << theToken.mLineNumber << "]";
      errMsg = out.str();
      return(NULL);
   }

   ++idx;

   if(idx >= n ||
      (t[idx].mTxt != PARSER_VOLUME && t[idx].mTxt != PARSER_INSTRUMENT))
   {
      return(
         new NoteNode(noteNumber, duration, volume, instrument)
      );
   }

   if((foundInstrument && t[idx].mTxt == PARSER_INSTRUMENT) ||
      (foundVolume     && t[idx].mTxt == PARSER_VOLUME))
   {
      out <<
         "Multiple " <<
         (foundVolume ? "volume" : "instrument") <<
         " settings, line: [" << t[idx].mLineNumber << "]";

      errMsg = out.str();

      return(NULL);
   }

   ++idx;
   theToken = t[idx];
   error    = false;

   if(foundVolume)
   {
      if(!getInstrumentNumber(theToken.mTxt, &instrument, errMsg))
         error = true;
   }
   else
   {
      if(!getVolume(theToken.mTxt, &volume, errMsg))
         error = true;
   }

   if(error)
   {
      out << errMsg << ", line: [" << theToken.mLineNumber << "]";
      errMsg = out.str();
      return(NULL);
   }

   idx++;

   return(new NoteNode(noteNumber, duration, volume, instrument));
}

static int operator < (NoteStateChange &a, NoteStateChange &b)
{
   int aValue = a.getTimeOffset();
   int bValue = b.getTimeOffset();

   if(aValue < bValue)
      return(true);

   if(aValue == bValue && a.getVolume() < b.getVolume())
      return(true);

   return(false);
}

static void dumpTokens(TokenVector &tokens)
{
   FOREACH(TokenVector, tokens, it)
   {
      cout << it->mLineNumber << ":" << it->mTxt << endl;
   }
}

static void dumpEvents(EventPtrVector &events)
{
   FOREACH(EventPtrVector, events, e)
      cout << (*e)->toString() << endl;
}

static void dumpNodes(NodePtrVector &nodes)
{
   FOREACH(NodePtrVector, nodes, it)
   {
      cout <<
         "Duration: " << (*it)->getDuration() << ", " <<
         "Node: " << (*it)->toString() << endl;
   }
}

static void dumpChanges(string caption, NoteStateChangeVector &changes)
{
   cout << caption << endl;

   FOREACH(NoteStateChangeVector, changes, it)
   {
      cout << it->getTimeOffset() << ": " << it->getNumber();

      if(it->isOn())
      {
         cout <<
            ", " << it->getVolume() <<
            ", " << it->getInstrument();
      }

      cout << endl;
   }
}
