// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Revision History
// ---------------- 
//
// Version 1.0 - Apr 10 2008 - Initial Release
//
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/*****************************************************************************
 * C/C++ Headers
 *****************************************************************************/

#include <iostream>
#include <iomanip>
#include <string>
#include <ios>

#include <sys/timeb.h>
#include <stdio.h>
#include <time.h>

/*****************************************************************************
 * Local Headers
 *****************************************************************************/

#include "cleaner.h"
#include "player.h"
#include "parser.h"
#include "midi.h"

/*****************************************************************************
 * Using Declarations
 *****************************************************************************/

using std::string;
using std::cout;
using std::endl;
using std::ios;
using std::setw;

/*****************************************************************************
 * Program Version
 *****************************************************************************/

static const char *VERSION = "Version 1.0 - Apr 10 2008";

/*****************************************************************************
 * Local Functions
 *****************************************************************************/

static double now()
{
   struct _timeb timebuffer;

   _ftime(&timebuffer);

   double value  = (double)timebuffer.time;
   double millis = (double)timebuffer.millitm;

   value += millis / 1000.0;

   return(value);
}

static void usage(const char *program)
{
   cout <<
      "Usage: " << program << " fileName " <<
      "( " << PARSER_VOLUME     << " defaultVolume ) " <<
      "( " << PARSER_INSTRUMENT << " defaultInstrument )" <<
      endl << endl << "Instruments:" << endl << endl;

   int maxLength = 0;

   for(int i = MIDI_MIN_INSTRUMENT; i <= MIDI_MAX_INSTRUMENT; i ++)
   {
      string name = Midi::instrumentNumberToName(i);
      int length  = name.size();

      if(length > maxLength)
         maxLength = length;
   }

   cout.setf(ios::left);

   for(int j = MIDI_MIN_INSTRUMENT; j <= MIDI_MAX_INSTRUMENT; j ++)
   {
      cout << "   " << setw(maxLength) <<
         Midi::instrumentNumberToName(j) << " " << j << endl;
   }

   exit(1);
}

/*****************************************************************************
 * Definition - Class: Player
 *****************************************************************************/

bool Player::play(
   const string &contents,
   int            defaultVolume,
   int            defaultInstrument,
   PlayerCallBack &theCallBack,
   string         &errMsg
)
{
   EventPtrVector events;
   int timeOffset = 0;

   bool status = Parser::getEvents(
      contents,
      defaultVolume,
      defaultInstrument,
      events,
      timeOffset,
      errMsg
   );

   if(!status)
      return(false);

   Cleaner theCleaner;

   theCleaner.addPointerVector(events);

   Midi *m = Midi::make(0, errMsg);

   if(!m)
      return(false);

   theCleaner.addPointer(m);

   double startTime = now();
   double totalTime = timeOffset / 1000.0;

   for(EventPtrVector::iterator i = events.begin(); i != events.end(); i ++)
   {
      (*i)->apply(m);

      double elapsed = now() - startTime;
      double pct = 100.0 * elapsed / totalTime;

      if(!theCallBack.process(pct > 100.0 ? 100.0 : pct))
         break;
   }

   return(true);
}

/*****************************************************************************
 * main
 *****************************************************************************/

int main(int argc, char **argv)
{
   if(argc < 2)
      usage(argv[0]);

   string instrument     = "acoustic_grand_piano";
   int defaultVolume     = 100;
   int defaultInstrument = Midi::instrumentNameToNumber(instrument);

   string errMsg;
   string fileName;

   bool status = Parser::getArgs(
      argc,
      argv,
      &defaultVolume,
      &defaultInstrument,
      fileName,
      errMsg
   );

   if(!status)
   {
      cout << errMsg << endl;
      return(1);
   }

   string contents;

   if(!Parser::fileToString(fileName, contents, errMsg))
   {
      cout << errMsg << endl;
      return(1);
   }

   PlayerCallBack theCallBack;

   status = Player::play(
      contents,
      defaultVolume,
      defaultInstrument,
      theCallBack,
      errMsg
   );

   if(!status)
   {
      cout << errMsg << endl;
      return(1);
   }

   return(0);
}
